#!/usr/bin/env python
"""
Lists collaborators for GitHub repo or repos starting with pattern.

Optionally with -xls use a spreadsheet filename and column to find Github usernames
that haven't accepted their invite yet (assuming you sent it previously).

    python ListGithubCollab.py ~/.ssh/oauth myowg -stem foo- -xls users.xlsx C

oauth token must have "read:org" (and "repo" for private repos) permissions
https://developer.github.com/v3/repos/#oauth-scope-requirements
"""
from argparse import ArgumentParser
import pandas
from pathlib import Path
import itertools

from pygithubutils import check_api_limit, connect_github, get_collabs


def main(P):
    op, sess = connect_github(P.oauth, P.orgname)
    if not check_api_limit(sess):
        raise RuntimeError("GitHub API limit exceeded")

    return get_collabs(op, sess, P.stem, P.regex)


if __name__ == "__main__":
    p = ArgumentParser()
    p.add_argument("oauth", help="Oauth filename")
    p.add_argument("orgname", help="Github organization name")
    p.add_argument("-stem", help="reponame starts with")
    p.add_argument("-regex", help="regex pattern of reponame")
    p.add_argument("-xls", help="spreadsheet filename and column to find missing usernames (who isn't signed up)", nargs=2)
    P = p.parse_args()

    collabs = main(P)

    for k in sorted(collabs.keys()):
        print(k, collabs[k])

    if P.xls:
        xlsfn = Path(P.xls[0]).expanduser()
        required = pandas.read_excel(xlsfn, usecols=P.xls[1]).squeeze().tolist()
        present = list(itertools.chain.from_iterable(collabs.values()))

        missing = set(required).difference(present)
        print("Missing:\n", missing)



#!/usr/bin/env python
"""
mass add team members to repos, optionally creating new repos

    python CreateGithubTeamRepos.py my.xlsx ~/.ssh/orgOauth -stem sw -orgname myorg -col B C

    python CreateGithubTeamRepos.py my.xlsx ~/.ssh/orgOauth -stem sw -orgname myorg -col B D E

oauth token must have "write:org" and public_repo (or repo for private) permissions
https://developer.github.com/v3/repos/#oauth-scope-requirements
"""
import pandas
from pygithubutils import repo_exists, check_api_limit, connect_github
from pathlib import Path
import warnings
from argparse import ArgumentParser

USERNAME = "GitHub"
TEAMS = "Team"
NAME = "Name"


def main():
    p = ArgumentParser()
    p.add_argument("fn", help=".xlsx with group info")
    p.add_argument("oauth", help="Oauth file")
    p.add_argument("-stem", help="beginning of repo names", default="")
    p.add_argument("-orgname", help="Github Organization", required=True)
    p.add_argument("-col", help="columns for Username, teamname", nargs="+", required=True)
    p.add_argument("-private", help="create private repos", action="store_true")
    p.add_argument("-create", help="create repo if not existing", action="store_true")
    p = p.parse_args()

    fn = Path(p.fn).expanduser()

    teams = pandas.read_excel(fn, usecols=",".join(p.col)).squeeze().dropna()
    if not teams.ndim == 2:
        raise ValueError("need to have member names and team names. Check that -col argument matches spreadsheet.")
    # %%
    op, sess = connect_github(p.oauth, p.orgname)
    if not check_api_limit(sess):
        raise RuntimeError("GitHub API limit exceeded")

    adder(teams, p.stem, p.private, p.create, op, sess)


def adder(teams: pandas.DataFrame, stem: str, private: bool, create: bool, op, sess):
    for _, row in teams.iterrows():
        if row.size == 3:
            reponame = f"{stem}{row[TEAMS]:02.0f}-{row[NAME]}"
        elif row.size == 2:
            reponame = f"{stem}{row[TEAMS]}"
        else:
            raise ValueError("I expect team number OR team number and team name")

        username = row[USERNAME]

        if create and not repo_exists(op, reponame):
            print("creating", reponame)
            op.create_repo(name=reponame, private=private)

        repo = op.get_repo(reponame)

        if not repo.has_in_collaborators(username):
            try:
                repo.add_to_collaborators(username)
                print(f"{username} invited to {reponame}")
            except Exception:
                warnings.warn(f"failed to invite {username} to {reponame}")


if __name__ == "__main__":
    main()



"""
How many total GitHub stars do you have?
"""
from typing import Tuple, List
from time import sleep
from pathlib import Path
import github
import logging

from .base import check_api_limit, github_session, get_repos, user_or_org


def repo_prober(
    user: str, oauth: Path = None, branch: str = None, starsonly: bool = False, verbose: bool = False
) -> Tuple[List[Tuple[str, int, int]], List[Tuple[str, int]]]:
    """
    probe all GitHub repos for a user to see how much forks of each repo are ahead.
    Discover if there is an actively developed fork of your GitHub repos

    Parameters
    ----------
    user : str
        GitHub username
    oauth : pathlib.Path, optional
        file containing GitHub Oauth hash
    branch : str, optional
        Git branch to examine
    starsonly: bool, optional
        far faster to only count forks and stars
    verbose : bool, optional
        verbosity

    Results
    -------
    counts : list of tuple of str, int, int
        forks and stars for each repo
    ahead : list of tuple of str, int
        forked with repos with number of commits they're ahead of your repo
    """
    # %% authentication
    sess = github_session(oauth)
    check_api_limit(sess)
    userorg = user_or_org(sess, user)
    # %% prepare to loop over repos
    repos = get_repos(userorg)

    counts: List[Tuple[str, int, int]] = []
    ahead: List[Tuple[str, int]] = []

    for repo in repos:
        if not starsonly:
            ahead += fork_prober(repo, sess, ahead, branch, verbose)

        counts.append((repo.name, repo.forks_count, repo.stargazers_count))  # type: ignore
        # FIXME: bug in PyGithub fixed by https://github.com/PyGithub/PyGithub/pull/1513

        check_api_limit(sess)

    return counts, ahead


def fork_prober(
    repo: github.Repository.Repository, sess: github.Github, ahead: List[Tuple[str, int]], branch: str = None, verbose: bool = False
) -> List[Tuple[str, int]]:
    """
    check a GitHub repo for forks

    Parameters
    ----------
    repo :
        handle to GitHub repo
    sess :
        handle to GitHub session
    ahead : list of tuple of str, int
        forked with repos with number of commits they're ahead of your repo
    branch : str, optional
        Git branch to examine
    verbose : bool, optional
        verbosity

    Results
    -------
    ahead : list of tuple of str, int
        forked with repos with number of commits they're ahead of your repo
    """
    check_api_limit(sess)

    b = repo.default_branch if not branch else branch

    try:
        master = repo.get_branch(b)
    except github.GithubException as e:
        logging.error(f"{repo.full_name}  {e}")
        return ahead

    forks = repo.get_forks()
    for fork in forks:
        sleep(0.1)  # don't hammer the API, avoiding 502 errors

        check_api_limit(sess)

        try:
            fmaster = fork.get_branch(b)
        except github.GithubException as e:
            if e.data["message"] == "Not Found":  # repo/branch that they deleted  FIXME: should we check their default branch?
                continue

            logging.error(f"{repo.full_name} {fork.full_name}  {e}")
            continue

        try:
            comp = repo.compare(master.commit.sha, fmaster.commit.sha)
        except github.GithubException as excp:
            # if excp.data["message"].startswith("No common ancestor"):
            #     continue

            logging.error(f"{repo.full_name} {fork.full_name}  {excp}")
            continue

        if comp.ahead_by:
            ahead.append((fork.full_name, comp.ahead_by))
            print(f"{fork.full_name} ahead by {comp.ahead_by}", end="")
            if verbose and comp.behind_by:
                print(f"behind by {comp.behind_by}", end="")
            print()

    return ahead



#!/usr/bin/env python
"""
Duplicate repos specified in spreadsheet.
Requires GitHub Oauth login.

The Oauth file should be in a secure place, NOT in a Git repo!
Maybe encrypted and with permissions 600.
The Oauth key must have "repo" checked, or you'll get 404 error on user.create_repo().

Assumes you have an SSH key loaded for git push --mirror step
"""
from argparse import ArgumentParser
import pygithubutils.duplicator as gu
import pygithubutils as gb


def main():
    p = ArgumentParser()
    p.add_argument("fn", help="spreadsheet filename")
    p.add_argument("oauth", help="Oauth filename")
    p.add_argument("-u", "--username", help="username or organization to create duplicate under")
    p.add_argument("-s", "--stem", help="beginning of duplicated repo names")
    p.add_argument("-w", "--sheet", help="excel sheet to process", required=True)
    p = p.parse_args()

    repos = gb.read_repos(p.fn, p.sheet)

    gu.repo_dupe(repos, p.oauth, p.username, p.stem)


if __name__ == "__main__":
    main()



"""
GitHub API utilities
"""
from pathlib import Path
from datetime import datetime
import logging
import typing as T
import pandas

import github

TYPE_USERORG = T.Union[github.NamedUser.NamedUser, github.Organization.Organization]
TYPE_AUTH = T.Union[github.AuthenticatedUser.AuthenticatedUser, github.Organization.Organization]


def check_api_limit(g: github.Github = None) -> bool:
    """
    https://developer.github.com/v3/#rate-limiting
    don't hammer the API, avoiding 502 errors

    No penalty for checking rate limits

    Parameters
    ----------
    g : optional
        GitHub session

    Results
    -------
    ok : bool
        haven't yet exceeded GitHub API limits
    """
    if g is None:
        g = github_session()

    ok = True

    api_limits = g.rate_limiting  # remaining, limit
    api_remaining, api_max = api_limits
    treset = datetime.utcfromtimestamp(g.rate_limiting_resettime)  # local time

    if api_remaining == 0:
        raise ConnectionRefusedError(f"GitHub rate limit exceeded: {api_remaining} / {api_max}. Try again after {treset} UTC.")
        ok = False
    elif api_remaining < 10:
        logging.warning(ResourceWarning(f"approaching GitHub API limit, {api_remaining} / {api_max} remaining until {treset} UTC."))
    else:
        logging.debug(f"GitHub API limit: {api_remaining} / {api_max} remaining until {treset} UTC.")

    return ok


def github_session(oauth: Path = None) -> github.Github:
    """
    setup GitHub session

    Parameters
    ----------

    oauth : pathlib.Path, optional
        file containing GitHub Oauth hash

    Results
    -------
    g : github.Github
        GitHub session handle
    """
    if oauth:
        oauth = Path(oauth).expanduser()
        g = github.Github(oauth.read_text().strip())  # no trailing \n allowed
    else:  # unauthenticated
        g = github.Github()

    return g


def connect_github(oauth: Path, orgname: str = None) -> T.Tuple[TYPE_AUTH, github.Github]:
    """
    retrieve organizations or users from GitHub

    Parameters
    ----------
    oauth : pathlib.Path
        file containing GitHub Oauth hash
    orgname : str
        organization name or username

    Results
    -------
    op : github.AuthenticatedUser.AuthenticatedUser or github.Organization.Organization
        handle to organization or user
    sess : github.Github
        GitHub session
    """
    sess = github_session(oauth)
    guser = sess.get_user()

    org = None
    if orgname:
        assert isinstance(guser, github.AuthenticatedUser.AuthenticatedUser)
        orgs = list(guser.get_orgs())
        for o in orgs:
            if o.login == orgname:
                org = o
                break

        if org is None:
            raise ValueError(f"Organization {org} authentication could not be established")
        op = org
    else:
        assert isinstance(guser, github.Organization.Organization)
        op = guser

    return op, sess


def repo_exists(user: TYPE_AUTH, reponame: str) -> bool:
    """
    Does a particular GitHub repo exist?

    Parameters
    ----------
    user : github.AuthenticatedUser.AuthenticatedUser or github.Organization.Organization
        GitHub user or organizaition handle
    reponame : str
        reponame under user e.g. pymap3d

    Results
    -------
    exists : bool
        GitHub repo exists
    """
    exists = False
    try:
        repo = user.get_repo(reponame)
        if repo.name:
            exists = True
    except github.GithubException as e:
        logging.info(str(e))
        pass

    return exists


def last_commit_date(sess: github.Github, name: str) -> datetime:
    """
    What is the last commit date to this repo.

    Equivalent to:

        git show -s --format=%cI HEAD


    Parameters
    ----------
    sess : github.Github
        GitHub session
    name : str
        name of GitHub repo e.g. pymap3d

    Results
    -------
    time : datetime.datetime
        time of last repo modification
    """
    time = None

    try:
        repo = sess.get_repo(name)
        if not repo_isempty(repo):
            time = repo.pushed_at
    except github.GithubException as e:
        logging.error(f"{name} not found \n")
        logging.info(str(e))

    return time


def repo_isempty(repo: github.Repository.Repository) -> bool:
    """
    is a GitHub repo empty?

    Parameters
    ----------
    repo : github.Repository
        handle to GitHub repo

    Results
    -------
    empty : bool
        GitHub repo empty
    """
    try:
        repo.get_contents("/")
        empty = False
    except github.GithubException as e:
        logging.error(f"{repo.name} is empty. \n")
        empty = True
        logging.info(str(e))

    return empty


def user_or_org(g: github.Github, user: str) -> T.Any:
    """
    Determines if user is a GitHub organization or standard user.
    This is relevant to getting private repos.

    Parameters
    ----------
    g: github.Github
        Github session handle
    user: str
        username or organization name

    Returns
    -------
    h: github.NamedUser.NamedUser or github.Organization.Organization
        the handle to the Organization or Username.
    """
    try:
        list(g.search_users(f"user:{user}"))
    except github.GithubException as e:
        raise ValueError(f"{user} not found on GitHub\n{e}")

    try:
        return g.get_organization(user)
    except github.GithubException:
        return g.get_user(user)


def read_repos(fn: Path, sheet: str) -> T.Dict[str, str]:
    """
    make pandas.Series of email/id, Git url from spreadsheet

    Parameters
    ----------
    fn : pathlib.Path
        path to Excel spreadsheet listing usernames and repos to duplicate
    sheet : str
        name of Excel sheet to use

    Results
    -------
    repos : dict
        all the repos to duplicate
    """

    # %% get list of repos to duplicate
    fn = Path(fn).expanduser()
    repos = pandas.read_excel(fn, sheet_name=sheet, index_col=0, usecols="A, D").squeeze()
    repos.dropna(how="any", inplace=True)

    return repos.to_dict()


def get_repos(userorg: TYPE_USERORG) -> T.Iterable[github.Repository.Repository]:
    """
    get list of Repositories for a user or organization

    Parameters
    ----------
    userorg: github.NamedUser.NamedUser or github.Organization.Organization
        username or organization handle

    Returns
    -------
    repos: list of github.Repository
        all repos for a username / orgname
    """
    return userorg.get_repos(type="all")



from time import sleep
from pathlib import Path
import subprocess
import logging
import tempfile
from datetime import datetime
import webbrowser
import shutil
import typing as T

import github

from .base import github_session, check_api_limit, last_commit_date, repo_exists

git = shutil.which("git")
if not git:
    raise ImportError("Git not found")


def repo_dupe(repos: T.Dict[str, str], oauth: Path, orgname: str = None, stem: str = ""):
    """
    Duplicate GitHub repos AND their wikis

    Parameters
    ----------
    repos: dict of str, str
        GitHub username, reponame to duplicate
    oauth: pathlib.Path
        GitHub Oauth token  https://github.com/settings/tokens
    orgname: str
        create repos under Organization instead of username
    stem: str
        what to start new repo name with
    """
    # %% authenticate
    sess = github_session(oauth)
    guser = sess.get_user()

    org = None
    if orgname:
        assert isinstance(guser, github.AuthenticatedUser.AuthenticatedUser)
        orgs = list(guser.get_orgs())
        for o in orgs:
            if o.login == orgname:
                org = o
                break

        assert org is not None
        op = org
    else:
        assert isinstance(guser, github.Organization.Organization)
        op = guser

    username = op.login

    if not check_api_limit(sess):
        raise RuntimeError("GitHub API limit exceeded")
    # %% prepare to loop over repos
    for email, oldurl in repos.items():
        if not check_api_limit(sess):
            raise RuntimeError("GitHub API limit exceeded")

        oldurl = oldurl.replace("https", "ssh")
        oldname = "/".join(oldurl.split("/")[-2:]).split(".")[0]

        oldtime = last_commit_date(sess, oldname)
        if oldtime is None:
            continue

        mirrorname = stem + email

        gitdupe(oldurl, oldtime, username, mirrorname, op)
        gitdupe(oldurl, None, username, mirrorname, op, iswiki=True)

        sleep(0.1)


def gitdupe(oldurl: str, oldtime: T.Optional[datetime], username: str, mirrorname: str, op, iswiki: bool = False):

    if iswiki:
        oldurl += ".wiki.git"
        mirrorname += ".wiki.git"
        try:
            subprocess.check_call([git, "ls-remote", "--exit-code", oldurl], stdout=subprocess.DEVNULL)
        except subprocess.CalledProcessError:
            logging.error(f"{oldurl} has no Wiki")
            return

    newname = f"{username}/{mirrorname}"
    newurl = f"ssh://github.com/{newname}"

    if not iswiki:
        exists = repo_exists(op, mirrorname)
        if exists:
            newrepo = op.get_repo(mirrorname)
            if newrepo.pushed_at >= oldtime:
                return

    else:
        try:
            subprocess.check_call([git, "ls-remote", "--exit-code", newurl], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
            return
        except subprocess.CalledProcessError:
            exists = True

    print("\n", oldurl, "\n")

    with tempfile.TemporaryDirectory() as d:
        tmprepo = Path(d)
        # 1. bare clone
        cmd = [git, "clone", oldurl] if iswiki else ["git", "clone", "--bare", oldurl]
        subprocess.check_call(cmd, stdout=subprocess.DEVNULL, cwd=tmprepo)

        # 2. create new repo
        if not exists:
            op.create_repo(name=mirrorname, private=True, has_wiki=True)

        # 3. mirror to new repo

        if iswiki:
            dupewiki(tmprepo, oldurl, newurl)
        else:
            pwd = tmprepo / (oldurl.split("/")[-1])
            pwd = pwd.with_suffix(".git")

            cmd = [git, "push", "--mirror", newurl]
            subprocess.check_call(cmd, cwd=pwd)


def dupewiki(prepo: Path, oldurl: str, newurl: str):
    """
    Note: GitLab API has Wiki included, but at this time, GitHub API does not cover Wiki
    """
    pwd = prepo / (oldurl.split("/")[-1]).split(".git")[0]

    subprocess.check_call([git, "remote", "set-url", "origin", newurl], cwd=pwd, stdout=subprocess.DEVNULL)

    browseurl = newurl
    browseurl = browseurl.replace("ssh", "https").replace(".wiki.git", "/wiki")
    webbrowser.open_new_tab(browseurl)
    sleep(10.0)  # TODO: use suprocess.run() instead of webbrowser

    subprocess.check_call([git, "push", "-f"], cwd=pwd)



#!/usr/bin/env python
"""
mass create repos for teams

example with spreadsheet with usernames in column C, teamname in column D

    python CreateGithubTeamRepos.py my.xlsx ~/.ssh/orgOauth myorg -stem sw -col C D

oauth token must have "admin:org" and public_repo (or repo for private) permissions
https://developer.github.com/v3/repos/#oauth-scope-requirements
"""
import pandas
from pygithubutils import connect_github, repo_exists, check_api_limit
from pathlib import Path
from argparse import ArgumentParser

TEAMS = "Team"
NAME = "Name"


def main():
    p = ArgumentParser()
    p.add_argument("fn", help=".xlsx with group info")
    p.add_argument("oauth", help="Oauth file")
    p.add_argument("orgname", help="Github Organization")
    p.add_argument("-stem", help="beginning of repo names", default="")
    p.add_argument("-col", help="column(s) for TeamName OR TeamNumber, TeamName", nargs="+", required=True)
    p.add_argument("-private", help="create private repos", action="store_true")
    p = p.parse_args()

    fn = Path(p.fn).expanduser()

    teams = pandas.read_excel(fn, usecols=",".join(p.col)).squeeze().dropna().drop_duplicates()
    # %%
    op, sess = connect_github(p.oauth, p.orgname)
    if not check_api_limit(sess):
        raise RuntimeError("GitHub API limit exceeded")

    if teams.ndim == 1:
        by_num(teams, p.stem, p.private, op, sess)
    elif teams.shape[1] == 2:
        by_name(teams, p.stem, p.private, op, sess)


def by_name(teams: pandas.DataFrame, stem: str, private: bool, op, sess):
    for _, row in teams.iterrows():
        reponame = f"{stem}{row[TEAMS]:02.0f}-{row[NAME]}"
        if repo_exists(op, reponame):
            continue

        print(f"creating {op.login}/{reponame}")
        op.create_repo(name=reponame, private=private)


def by_num(teams: pandas.DataFrame, stem: str, private: bool, op, sess):
    for teamnum in teams.values:
        reponame = f"{stem}{teamnum}"
        if repo_exists(op, reponame):
            continue

        print("creating", reponame)
        op.create_repo(name=reponame, private=private)


if __name__ == "__main__":
    main()



import typing
import re

import github


def get_collabs(
    op: github.Organization.Organization, sess: github.Github, stem: str = None, regex: str = None
) -> typing.Dict[str, typing.List[str]]:
    """
    get collaborators of a GitHub repo

    Parameters
    ----------

    stem: str, optional
        repo names starts with
    regex: str, optional
        regex pattern

    Return
    ------

    collabs: dict of list of str
        collaborators on all repos selected in organization
    """
    orgmembers = [u.login for u in op.get_members()]

    collabs = {}
    rpat = re.compile(regex) if regex else None

    for repo in op.get_repos():
        if rpat and not rpat.match(repo.name):
            continue
        elif stem and not repo.name.startswith(stem):
            continue

        collabs[repo.name] = [u.login for u in repo.get_collaborators() if u.login not in orgmembers]

    return collabs
